@@ -1,7 +1,7 @@
 package Net::SAML2::Protocol::Assertion;
 use Moose;
 
-our $VERSION = '0.69'; # VERSION
+our $VERSION = '0.73'; # VERSION
 
 use MooseX::Types::DateTime qw/ DateTime /;
 use MooseX::Types::Common::String qw/ NonEmptySimpleStr /;
@@ -18,7 +18,7 @@ use Carp qw(croak);
 
 with 'Net::SAML2::Role::ProtocolMessage';
 
-# ABSTRACT: Net::SAML2::Protocol::Assertion - SAML2 assertion object
+# ABSTRACT: SAML2 assertion object
 
 
 has 'attributes' => (isa => 'HashRef[ArrayRef]', is => 'ro', required => 1);
@@ -37,66 +37,93 @@ has 'nameid_object' => (
     init_arg  => 'nameid',
     predicate => 'has_nameid',
 );
+has 'authnstatement_object' => (
+    isa       => 'XML::LibXML::Element',
+    is        => 'ro',
+    required  => 0,
+    init_arg  => 'authnstatement',
+    predicate => 'has_authnstatement',
+);
 
 
 
+sub _verify_encrypted_assertion {
+    my $self     = shift;
+    my $xml      = shift;
+    my $cacert   = shift;
+    my $key_file = shift;
+    my $key_name = shift;
+
+    my $xpath = XML::LibXML::XPathContext->new($xml);
+    $xpath->registerNs('saml',  'urn:oasis:names:tc:SAML:2.0:assertion');
+    $xpath->registerNs('samlp', 'urn:oasis:names:tc:SAML:2.0:protocol');
+    $xpath->registerNs('dsig',  'http://www.w3.org/2000/09/xmldsig#');
+    $xpath->registerNs('xenc',  'http://www.w3.org/2001/04/xmlenc#');
+
+    return $xml unless $xpath->exists('//saml:EncryptedAssertion');
+
+    croak "Encrypted Assertions require key_file" if !defined $key_file;
+
+    $xml = $self->_decrypt(
+        $xml,
+        key_file => $key_file,
+        key_name => $key_name,
+    );
+    $xpath->setContextNode($xml);
+
+    my $assert_nodes = $xpath->findnodes('//saml:Assertion');
+    return $xml unless $assert_nodes->size;
+    my $assert = $assert_nodes->get_node(1);
+
+    return $xml unless $xpath->exists('dsig:Signature', $assert);
+    my $xml_opts->{ no_xml_declaration } = 1;
+    my $x   = Net::SAML2::XML::Sig->new($xml_opts);
+    my $ret = $x->verify($assert->toString());
+    die "Decrypted Assertion signature check failed" unless $ret;
+
+    return $xml unless $cacert;
+    my $cert = $x->signer_cert;
+    die "Certificate not provided in SAML Response, cannot validate" unless $cert;
+
+    my $ca = Crypt::OpenSSL::Verify->new($cacert, { strict_certs => 0 });
+    die "Unable to verify signer cert with cacert: " . $cert->subject
+        unless $ca->verify($cert);
+    return $xml;
+}
+
 sub new_from_xml {
     my($class, %args) = @_;
 
-    my $dom = no_comments($args{xml});
     my $key_file = $args{key_file};
-    my $cacert = $args{cacert};
+    my $cacert   = delete $args{cacert};
 
-    my $xpath = XML::LibXML::XPathContext->new($dom);
+    my $xpath = XML::LibXML::XPathContext->new();
     $xpath->registerNs('saml',  'urn:oasis:names:tc:SAML:2.0:assertion');
     $xpath->registerNs('samlp', 'urn:oasis:names:tc:SAML:2.0:protocol');
-    $xpath->registerNs('xenc', 'http://www.w3.org/2001/04/xmlenc#');
+    $xpath->registerNs('dsig',  'http://www.w3.org/2000/09/xmldsig#');
+    $xpath->registerNs('xenc',  'http://www.w3.org/2001/04/xmlenc#');
+
+    my $xml = no_comments($args{xml});
+    $xpath->setContextNode($xml);
+
+    $xml = $class->_verify_encrypted_assertion(
+        $xml,
+        $cacert,
+        $key_file,
+        $args{key_name},
+    );
+
+    my $dec = $class->_decrypt(
+        $xml,
+        key_file => $key_file,
+        key_name => $args{key_name}
+    );
+    $xpath->setContextNode($dec);
 
     my $attributes = {};
-
-    if ($xpath->findnodes('//saml:EncryptedAssertion')) {
-        if ( ! defined $key_file) {
-            die "Encrypted Assertions require key_file";
-        }
-        my $decrypted;
-        my $enc = XML::Enc->new(
-                        { key => $key_file , no_xml_declaration => 1 }, );
-        $decrypted  = $enc->decrypt($dom->toString());
-        $dom        = XML::LibXML->load_xml(string => $decrypted);
-        $xpath      = XML::LibXML::XPathContext->new($dom);
-        $xpath->registerNs('saml',  'urn:oasis:names:tc:SAML:2.0:assertion');
-        $xpath->registerNs('samlp', 'urn:oasis:names:tc:SAML:2.0:protocol');
-        $xpath->registerNs('dsig', 'http://www.w3.org/2000/09/xmldsig#');
-        $xpath->registerNs('xenc', 'http://www.w3.org/2001/04/xmlenc#');
-
-        my $xml_opts->{ no_xml_declaration } = 1;
-
-        my $assert = $xpath->findnodes('//saml:Assertion')->[0];
-        my @signedinfo = $xpath->findnodes('dsig:Signature', $assert);
-
-        if (defined $assert && (scalar @signedinfo ne 0)) {
-            my $x   = Net::SAML2::XML::Sig->new($xml_opts);
-            my $ret = $x->verify($assert->serialize);
-            die "Decrypted Assertion signature check failed" unless $ret;
-
-            if ($cacert) {
-                my $cert = $x->signer_cert
-                    or die "Certificate not provided and not in SAML Response, cannot validate";
-
-                my $ca = Crypt::OpenSSL::Verify->new($cacert, { strict_certs => 0, });
-                if (! $ca->verify($cert)) {
-                    die "Decrypted Assertion - Unable to verify signer cert with cacert: $cert->subject";
-                }
-            }
-        }
-    }
-
-    for my $node (
-        $xpath->findnodes('//saml:Assertion/saml:AttributeStatement/saml:Attribute'))
+    for my $node ($xpath->findnodes('//saml:Assertion/saml:AttributeStatement/saml:Attribute/saml:AttributeValue/..'))
     {
-        # We can't select by saml:AttributeValue
-        # because of https://rt.cpan.org/Public/Bug/Display.html?id=8784
-        my @values = $node->findnodes("*[local-name()='AttributeValue']");
+        my @values = $xpath->findnodes("saml:AttributeValue", $node);
         $attributes->{$node->getAttribute('Name')} = [map $_->string_value, @values];
     }
 
@@ -125,13 +152,18 @@ sub new_from_xml {
     }
 
     my $nameid;
-    if (my $node = $xpath->findnodes('//samlp:Response/saml:Assertion/saml:Subject/    saml:NameID')) {
+    if (my $node = $xpath->findnodes('/samlp:Response/saml:Assertion/saml:Subject/saml:NameID')) {
         $nameid = $node->get_node(1);
     }
     elsif (my $global = $xpath->findnodes('//saml:Subject/saml:NameID')) {
         $nameid = $global->get_node(1);
     }
 
+    my $authnstatement;
+    if (my $node = $xpath->findnodes('/samlp:Response/saml:Assertion/saml:AuthnStatement')) {
+        $authnstatement = $node->get_node(1);
+    }
+
     my $nodeset = $xpath->findnodes('/samlp:Response/samlp:Status/samlp:StatusCode|/samlp:ArtifactResponse/samlp:Status/samlp:StatusCode');
 
     croak("Unable to parse status from assertion") unless $nodeset->size;
@@ -158,12 +190,14 @@ sub new_from_xml {
         in_response_to => $xpath->findvalue('//saml:Subject/saml:SubjectConfirmation/saml:SubjectConfirmationData/@InResponseTo'),
         response_status => $status,
         $sub_status ? (response_substatus => $sub_status) : (),
+        $authnstatement ? (authnstatement => $authnstatement) : (),
     );
 
     return $self;
 }
 
 
+
 sub name {
     my $self = shift;
     return $self->attributes->{CN}[0];
@@ -205,6 +239,91 @@ sub nameid_sp_provided_id {
 }
 
 
+sub authnstatement {
+    my $self = shift;
+    return unless $self->has_authnstatement;
+    return $self->authnstatement_object->textContent;
+}
+
+
+sub authnstatement_authninstant {
+    my $self = shift;
+    return unless $self->has_authnstatement;
+    return $self->authnstatement_object->getAttribute('AuthnInstant');
+}
+
+
+sub authnstatement_sessionindex {
+    my $self = shift;
+    return unless $self->has_authnstatement;
+    return $self->authnstatement_object->getAttribute('SessionIndex');
+}
+
+
+sub authnstatement_subjectlocality {
+    my $self = shift;
+    return unless $self->has_authnstatement;
+
+    my $xpc = XML::LibXML::XPathContext->new;
+    $xpc->registerNs('saml',  'urn:oasis:names:tc:SAML:2.0:assertion');
+    my $subjectlocality;
+    my $xpath_base = '//saml:AuthnStatement/saml:SubjectLocality';
+    if (my $nodes = $xpc->find($xpath_base, $self->authnstatement_object)) {
+        my $node = $nodes->get_node(1);
+        $subjectlocality = $node;
+    }
+    return $subjectlocality;
+}
+
+
+sub subjectlocality_address {
+    my $self = shift;
+    return unless $self->has_authnstatement;
+    my $subjectlocality = $self->authnstatement_subjectlocality;
+    return unless $subjectlocality;
+    return $subjectlocality->getAttribute('Address');
+}
+
+
+sub subjectlocality_dnsname {
+    my $self = shift;
+    return unless $self->has_authnstatement;
+    my $subjectlocality = $self->authnstatement_subjectlocality;
+    return unless $subjectlocality;
+    return $subjectlocality->getAttribute('DNSName');
+}
+
+
+sub authnstatement_authncontext {
+    my $self = shift;
+    return unless $self->has_authnstatement;
+
+    my $xpc = XML::LibXML::XPathContext->new;
+    $xpc->registerNs('saml',  'urn:oasis:names:tc:SAML:2.0:assertion');
+    my $authncontext;
+    my $xpath_base = '//saml:AuthnStatement/saml:AuthnContext';
+    if (my $nodes = $xpc->find($xpath_base, $self->authnstatement_object)) {
+        my $node = $nodes->get_node(1);
+        $authncontext = $node;
+    }
+    return $authncontext;
+}
+
+
+sub contextclass_authncontextclassref {
+    my $self = shift;
+    return unless $self->has_authnstatement;
+    my $authncontextclassref = $self->authnstatement_authncontext;
+    return unless $authncontextclassref;
+    my $xpc = XML::LibXML::XPathContext->new;
+    $xpc->registerNs('saml',  'urn:oasis:names:tc:SAML:2.0:assertion');
+    if (my $value = $xpc->findvalue('//saml:AuthnContextClassRef', $self->authnstatement_object)) {
+        $authncontextclassref = $value;
+    }
+    return $authncontextclassref;
+}
+
+
 sub valid {
     my ($self, $audience, $in_response_to) = @_;
 
@@ -231,6 +350,22 @@ sub success {
     return 0;
 }
 
+sub _decrypt {
+    my $self    = shift;
+    my $xml     = shift;
+    my %options = @_;
+
+    return $xml unless $options{key_file};
+
+    my $enc = XML::Enc->new(
+        {
+            no_xml_declaration => 1,
+            key                => $options{key_file},
+        }
+    );
+    return XML::LibXML->load_xml(string => $enc->decrypt($xml, %options));
+}
+
 1;
 
 __END__
@@ -241,11 +376,11 @@ __END__
 
 =head1 NAME
 
-Net::SAML2::Protocol::Assertion - Net::SAML2::Protocol::Assertion - SAML2 assertion object
+Net::SAML2::Protocol::Assertion - SAML2 assertion object
 
 =head1 VERSION
 
-version 0.69
+version 0.73
 
 =head1 SYNOPSIS
 
@@ -324,6 +459,38 @@ Returns the NameID SPNameQualifier
 
 Returns the NameID SPProvidedID
 
+=head2 authnstatement
+
+Returns the AuthnStatement
+
+=head2 authnstatement_authninstant
+
+Returns the AuthnStatement AuthnInstant
+
+=head2 authnstatement_sessionindex
+
+Returns the AuthnStatement SessionIndex
+
+=head2 authnstatement_subjectlocality
+
+Returns the AuthnStatement SubjectLocality
+
+=head2 subjectlocality_address
+
+Returns the SubjectLocality Address
+
+=head2 subjectlocality_dnsname
+
+Returns the SubjectLocality DNSName
+
+=head2 authnstatement_authncontext
+
+Returns the AuthnContext for the AuthnStatement
+
+=head2 contextclass_authncontextclassref
+
+Returns the ContextClass AuthnContextClassRef
+
 =head2 valid( $audience, $in_response_to )
 
 Returns true if this Assertion is currently valid for the given audience.
